// $Ik: env.C,v 1.5 2002/03/02 12:49:52 dvermeir Exp $
#include <fstream>
#include <iterator>
#include <algorithm>

#include <dvutil/ntostr.h>

#include "env.h"

using namespace std;

namespace Dv {
namespace Util {

//////////////////////////////////////////////////////////////////////
////////////// Dv::Util::Env
//////////////////////////////////////////////////////////////////////

const string Env::Exception::NAME("Dv::Util::Env::Exception");

Env::Env(const Env& e): box_(0) {
merge(e);
}

Env::Env(istream& is) throw (Exception): box_(0) {
read(is);
}

Env::Env(const string& fn) throw (Exception): box_(0) {
ifstream is(fn.c_str());
if (!is)
  throw Exception(fn + ": cannot open to read Dv::Util::Env");
read(is);
}

const Env&
Env::nil() {
static const Env nil_;
return nil_;
}

Env::Env(): box_(0) {
}

Env::~Env() {
delete box_;
}

Env::Type
Env::type() const {
if (!defined())
  return NIL;
return box_->type();
}

const char*
Env::type_s(Type t) {
static const char* types[] = 
	{ "nil", "sbox", "mbox", "abox" };
if (static_cast<size_t>(t) >= sizeof(types)/sizeof(const char*))
  return "unknown";
else
  return types[t];
}

size_t
Env::size() const {
switch (type()) {
  case NIL:
    return 0;
  case STRING:
    return 1;
  case MAP:
    return dynamic_cast<const MBox*>(box_)->size();
  case ARRAY:
    return dynamic_cast<const ABox*>(box_)->size();
  }
throw Exception("Env::size() failed, cannot happen");
}

const string&
Env::svalue() const throw (Exception) {
if (!box_)
  throw Exception("empty env in Env::svalue()");
return box_->svalue(); 
}

string&
Env::slvalue() throw (Exception) {
if (!box_)
  box_ = new SBox();
return box_->slvalue();
}

Env&
Env::operator[](int i) throw (Exception) { 
if (!box_)
  box_ = new ABox(i+1);
return *(*box_)[i];
}

const Env&
Env::operator()(int i) const throw (Exception) { 
if (!box_)
  return nil();
const Env* e = (*box_)(i);
if (e)
  return *e;
else
  return nil();
// return *(*box_)[i];
}

Env&
Env::operator[](const string& s) throw (Exception) {
if (!box_)
  box_ = new MBox();
return *(*box_)[s];
}

const Env&
Env::operator()(const string& s) const throw (Exception) { 
if (!box_)
  return nil();
const Env* e = (*box_)(s);
if (e)
  return *e;
else
  return nil();
// return *(*box_)(s);
}

Env&
Env::operator=(const string& s) { 
if (this!=&nil())
  slvalue() = s;
return *this;
}

Env&
Env::operator=(const Env& e) {
if (this == &e)
  return *this;
// clear old value
delete box_;
box_ = 0;
merge(e);
return *this;
}

Env&
Env::merge(const Env& e) {
if (this == &e)
  return *this;
switch (e.type()) {
 case NIL:
   break;
 case STRING:
   if (type() == NIL)
      slvalue() = e.svalue();
   break;
 case ARRAY: {
   Type t(type());
   if (t == ARRAY || t == NIL ) {
     size_t base = size();
     for (size_t i = 0; i < e.size(); ++i )
       (*this)[base+i] = e(i);
     }
   }
   break;
 case MAP: {
   Type t(type());
   if (t == MAP || t == NIL ) {
     for (iterator i=e.begin(); i!=e.end(); ++i)
       if (!(*this)((*i).first).defined())
         (*this)[(*i).first] = *((*i).second);
     }
   }
   break;
 default:
   throw Exception("Env::merge() failed, cannot happen");
 }
return *this;
}

bool
Env::operator==(const Dv::Util::Env& e) const {
if (size() != e.size())
  return false;
switch (type()) {
  case NIL:
    return true;
  case STRING:
    return svalue() == e.svalue();
  default:
    iterator i(begin());
    iterator j(e.begin());
    for (; i!=end() && j!=e.end() ; ++i, ++j) 
      if (!(  (*i).first == (*j).first &&
      	     *(*i).second == *(*j).second ))
        return false;
    return true;
  }
}

const Env*
Env::access(const string& key) const {
if (!defined())
  return 0;
static const string digits("0123456789");
if (key.size() && key.find_first_not_of(digits)==string::npos) { // integer
  long l;
  if (!cstr2long(key.c_str(), l))
    throw Exception(key + ": cstr2long failed -- should not happen");
  int i = static_cast<int>(l);
  if (is_array()) 
    return (*box_)(i);
  }
else 
  if (is_map()) 
    return (*box_)(key);
return 0;
}

static char
copy_until_matching(string& out, char left, char right, istream& is)
  throw (Env::Exception) {
/*
 * states:
 * 	0 = normal text
 *	1 = escape next char
 */
static const string 
  message(string("no matching '") + right + "' found");
int level = 1;
int state = 0;
char c;
char last_c = EOF;
while (is) {
 c = is.get();
 if (c == EOF)
   throw Env::Exception(message);
 switch (state) {
   case 0:
     if ( c=='\\' && (is.peek() == left || is.peek() == right) ) 
       state = 1;
     else if (c==left)
       ++ level;
     else if (c==right) {
       -- level;
       if (level == 0)
         return c;
       }
     out += c;
     break;
   case 1:
     out += c;
     break;
   }
 last_c = c;
 }
throw Env::Exception(message);
}

const Env&
Env::substitute(ostream& os, istream& is) const throw (Exception) {
// cerr << ">>> e = " << endl << *this << "<<<" << endl;
Display d(*this);
d.substitute(os,is);
return *this;
}

const Env&
Env::substitute(string& out, const string& in) const throw (Exception) {
ostringstream oss;
istringstream iss(in);
const Env& e(substitute(oss,iss));
out = oss.str();
return e;
}

bool
Env::explode(const string& s, vector<string>& keys) {
bool loop(false);
string::size_type start(0);
string::size_type end(string::npos);
while (start < s.size() && (end = s.find("::", start)) != string::npos) {
  string key(s, start, end - start);
  if (key.size()) {
    loop = (key == "*");
    if (!loop)
      keys.push_back(key);
    }
  start = end+2;
  }
if (start < s.size()) {
  string key(s,start);
  loop = (key == "*");
  if (!loop)
    keys.push_back(key);
  }
return loop;
}


Env&
Env::read_key(const string& s) throw (Exception) {
static const string digits("0123456789");
Env* e=this;
istringstream iss(s);
string key;
while (getline(iss,key,':')) {
  if (iss.peek() == ':')
    iss.ignore();
  // cerr << "read_key: key = " << key << endl;
  if (key.size()) {
    if (key.find_first_not_of(digits)==string::npos) { // integer => array
      long l;
      if (!cstr2long(key.c_str(), l))
        throw Exception(s + ": cstr2long failed -- should not happen");
      int i = static_cast<int>(l);
      e = &((*e)[i]);
      }
    else { // map
      e = &((*e)[key]);
      }
    }
  // else string, ok
  }
return *e;
}

Env::iterator 
Env::begin() const { 
if (defined() )
  return iterator(box_->begin() );
else
  return iterator(0);
// throw Exception("No iterator defined on this Env"); 
}

Env::iterator
Env::end() const { 
if (defined() )
  return iterator( box_->end() );
else
  return iterator(0);
// throw Exception("No iterator defined on this Env"); 
}

ostream&
Env::print(ostream& os, const string& key) const {
if (!defined())
  return os;
if (is_string()) {
  os << key << "=";
  const string& s(svalue());
  for (unsigned int j=0;(j<s.size());++j) {
    os << s[j];
    if (s[j]=='\n')
      os << '>';
    }
  return os << endl;
  }
// must be array of map
for (iterator i = begin(); i!=end(); ++i) {
  string newkey(key);
  if (newkey.size())
    newkey += "::";
  newkey += (*i).first;
  (*i).second->print(os, newkey);
  }
return os;
}

//////////////////////////////////////////////////////////////////////
////////////// Dv::Util::Env::Display
//////////////////////////////////////////////////////////////////////
Env::Display::Display(const Env& e): envs_(1, &e) {
}

Env::Display::Display(const Display& d): envs_(d.envs_), names_(d.names_) {
}

Env::Display& 
Env::Display::push(const string& key) {
/*****
const Env* e = top()->access(key);
*****/
const Env* e = (*this)(key);
if (e) {
  envs_.push_front(e);
  names_.push_front(key);
  }
else
  throw Exception(key + " not found in " + tostring(*this));
return *this;
}

Env::Display&
Env::Display::pop() {
if (size()>1) {
  envs_.pop_front();
  names_.pop_front();
  }
else
  throw Exception("pop from display of size " + tostring(size()));
return *this;
}

const Env*
Env::Display::operator()(const string& key) const {
const Env* e(0);
for (list<const Env*>::const_iterator i = envs_.begin(); i != envs_.end(); ++i) {
  e = (*i)->access(key);
  if (e) 
    return e;
  }
return 0;
}

const Env*
Env::Display::operator()(const vector<string>& keys) const {
if (keys.size() == 0)
  return top();
const Env* e = (*this)(keys[0]);
for (size_t i=1; e && i<keys.size(); ++i) 
  e = e->access(keys[i]);
return e;
}

Env::Display&
Env::Display::expand(const vector<string>& keys) {
try {
  if (keys.size() == 0)
    return *this;
  /*****
      relax, we now allow arbitray pushing of environments
      in a display 
  string key(keys[0]);
  while (top()->access(key) == 0)
    pop();
  // cerr << "after popping: " << *this << endl;
  push(key);
  for (size_t i=1; i<keys.size(); ++i)
    push(keys[i]);
  *****/
  for (size_t i=0; i<keys.size(); ++i)
    push(keys[i]);
  }
catch (exception& e) {
  cerr << "Display::expand"
       << "substitute: keys = ";
  copy(keys.begin(), keys.end(), ostream_iterator<string>(cerr," "));
  cerr << endl;
  cerr << e.what();
  throw;
  }
return *this;
}

ostream&
Env::Display::print(ostream& os) const {
copy(names_.begin(), names_.end(), ostream_iterator<string>(os,"."));
os << "<top>";
return os;
}

void
Env::Display::substitute(ostream& os, istream& is) const throw (Exception) {
if (size() == 0)
  throw Exception("Env::substitute: empty display");
// cerr << "substitute in " << *this << endl;
char c; // current char
char last_c = EOF; // char before current
while (is) {
  c = is.get();
  if (c==EOF)
    break;
  if (c=='$' && last_c != '\\' && is.peek() == '{') {
    is.ignore(); // skip '{'
    string index;
    getline(is, index, '}');
    // cerr << "substitute: index = " << index << endl;
    // c = (is ?  '}' : EOF); // last char read must be this or EOF
    vector<string> keys;
    bool loop = explode(index,keys); // loop if last key was '*'
    // cerr << "substitute: keys = ";
    // copy(keys.begin(), keys.end(), ostream_iterator<string>(cerr," "));
    // cerr << endl;
    const Env* ev = (*this)(keys);
    // cerr << "keys found => " << (ev ? "yes": "no") << endl;
    if (ev) { // found
      // cerr << ">>> ev " << endl << *ev << "<<< ev" << endl;
      if (is.peek() == '{') { // update scope
        is.ignore();
        string row;
        c = copy_until_matching(row, '{', '}', is);
	Display dnew(*this);
	dnew.expand(keys);
        if (loop) 
  	  for (Env::iterator i=ev->begin(); i!=ev->end(); ++i) {
	    // cerr << "pushing " << (*i).first << endl;
  	    dnew.push((*i).first);
            istringstream iss(row);
            dnew.substitute(os, iss);
	    // cerr << "popping " << endl;
            dnew.pop();
  	    }
        else {
          istringstream iss(row);
          dnew.substitute(os, iss);
          }
        }
      else { // simple substitution
	// cerr << "simple subst" << endl;
        if (ev->is_string() ) {
	  // cerr << "and string in simple subst" << endl;
          istringstream iss(ev->svalue());
          substitute(os, iss);
	  // cerr << "simple subst done" << endl;
          }
	}
      }
    else { // index not found
      cerr << "not found: " << index << " in " << *this << endl;
      os << "${" << index << "}";
      }
    }
  else 
    os.put(c);
  last_c = c;
  }
}


//////////////////////////////////////////////////////////////////////
////////////// Dv::Util::Env::iterator
//////////////////////////////////////////////////////////////////////

Env::iterator::value_type
Env::iterator::operator*() const { 
return **it_;
}
//////////////////////////////////////////////////////////////////////
////////////// Dv::Util::Env::Box
//////////////////////////////////////////////////////////////////////

Env::Type
Env::Box::type() const {
if ( dynamic_cast<const SBox*>(this) )
  return STRING;
if ( dynamic_cast<const MBox*>(this) )
  return MAP;
if ( dynamic_cast<const ABox*>(this) )
  return ARRAY;
throw Exception("Env::Box::type() failed, cannot happen");
}

Env::Box::iterator*
Env::Box::begin() const throw (Exception) {
return 0;
// throw Exception("No iterators defined"); 
}

Env::Box::iterator* 
Env::Box::end() const throw (Exception) { 
return 0;
// throw Exception("No iterators defined"); 
}

const string&
Env::Box::svalue() const throw (Exception) { 
throw Exception("no SBox in Box::svalue()"); 
}

string&
Env::Box::slvalue() throw (Exception) { 
throw Exception("no SBox in Box::slvalue()"); 
}

Env*
Env::Box::operator[](int i) throw (Exception) { 
throw Exception(Dv::Util::tostring(i) + ": no ABox but "
	+ Env::type_s(type()) + " in Box::operator[](int)"); 
}

const Env*
Env::Box::operator()(int i) const throw (Exception) { 
return &nil();
}

Env*
Env::Box::operator[](const string& s) throw (Exception) { 
throw Exception(s + ": no MBox but " + Env::type_s(type()) 
                  + " in Box::operator[](const string&)"); 
}

const Env*
Env::Box::operator()(const string& s) const throw (Exception) { 
return &nil();
}

Env::Box::~Box() {
}

//////////////////////////////////////////////////////////////////////
////////////// Dv::Util::Env::Box::iterator
//////////////////////////////////////////////////////////////////////

Env::Box::iterator&
Env::Box::iterator::operator++() {
throw Exception("operator++: no iterator defined on this Box");
}

Env::Box::iterator::value_type 
Env::Box::iterator::operator*() const {
throw Exception("operator*: no iterator defined on this Box");
}

bool
Env::Box::iterator::equal(const iterator* i) const {
throw Exception("equal: no iterator defined on this Box");
}

//////////////////////////////////////////////////////////////////////
////////////// Dv::Util::Env::ABox
//////////////////////////////////////////////////////////////////////

Env::ABox::ABox(size_t size): a_(size,0) {
for (size_t i = 0; i <= size; ++i)
  a_[i] = new Env();
}

Env::ABox::~ABox() {
for (vector<Env*>::const_iterator i = a_.begin(); i != a_.end(); ++i)
  delete *i;
}

Env*
Env::ABox::operator[](int i) throw (Exception) {
if (i<0)
  throw Exception(tostring(i) + ": bad index in ABox::operator[]");
if (static_cast<size_t>(i) >= a_.size()) {
  size_t old_size(a_.size());
  a_.resize(i+1,0);
  for (size_t j = old_size; j <= static_cast<size_t>(i); ++j)
    a_[j] = new Env();
  }
return a_[i];
}

const Env*
Env::ABox::operator()(int i) const throw (Exception) {
if (i<0)
  throw Exception(tostring(i) + ": bad index in ABox::operator()");
if (static_cast<size_t>(i) >= a_.size())
  return 0;
return a_[i];
}

//////////////////////////////////////////////////////////////////////
////////////// Dv::Util::Env::ABox::iterator
//////////////////////////////////////////////////////////////////////

Env::ABox::iterator::iterator(Vector::const_iterator it, 
  const Vector* v): it_(it), v_(v) {
}

Env::ABox::iterator::value_type 
Env::ABox::iterator::operator*() const { 
return make_pair(Util::tostring(it_ - v_->begin()), *it_); 
}

bool
Env::ABox::iterator::equal(const Box::iterator* i) const {
const iterator* j = dynamic_cast<const iterator*>(i);
if (j)
  return it_ == j->it_; 
return false;
}

//////////////////////////////////////////////////////////////////////
////////////// Dv::Util::Env::SBox
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
////////////// Dv::Util::Env::MBox
//////////////////////////////////////////////////////////////////////

Env::MBox::MBox() {
}

Env::MBox::~MBox() {
for (Map::const_iterator i = m_.begin(); i != m_.end(); ++i)
  delete (*i).second;
}

Env*
Env::MBox::operator[](const string& s) throw (Exception) { 
Map::const_iterator i = m_.find(s);
if (i==m_.end())
  return m_[s] = new Env();
return (*i).second;
}

const Env*
Env::MBox::operator()(const string& s) const throw (Exception) { 
Map::const_iterator i = m_.find(s);
if (i==m_.end())
  return 0;
  // throw Exception(s + ": bad key in MBox::operator()");
return (*i).second;
}

//////////////////////////////////////////////////////////////////////
////////////// Dv::Util::Env::MBox::iterator
//////////////////////////////////////////////////////////////////////

bool
Env::MBox::iterator::equal(const Box::iterator* i) const {
const iterator* j = dynamic_cast<const iterator*>(i);
if (j)
  return it_ == j->it_; 
return false;
}


//////////////////////////////////////////////////////////////////////
////////////// operator<<(ostream&, const Env&)
////////////// operator>>(istream&, Env&)
//////////////////////////////////////////////////////////////////////
Env&
Env::read(istream& is) throw (Dv::Util::Env::Exception) {
char c;
while (is) {
  // Assume we are at the beginning of a line, i.e. right after a '\n'
  // Skip comments and white space.
  bool skipping(true);
  while (skipping) {
    c = is.peek();
    if (c=='#') {
      string comment;
      getline(is,comment);
      }
    else if (isspace(c)) // too liberal, will also skip space in the
      is >> ws; // beginning of a line
    else
      skipping = false;
    }
  // Read a key.
  string key("");
  getline(is,key,'=');
  Dv::Util::Env& target(read_key(key));
  if (!is)
    return *this; 
  ostringstream  oss;
  // read value
  while (is.get(c))
    if (c=='\n') {
      if (is.get(c))
        if (c=='>')
          oss << '\n';
        else {
          is.putback(c);
          target = oss.str();
          break;
          }
      else {
        target = oss.str();
        break;
        }
      }
    else
      oss << c;
  }
return *this;
}

}}

namespace std {
ostream&
operator<<(ostream& os, const Dv::Util::Env& e) {
if (!e.defined())
  return os;
if (e.is_string())
  return os << e.svalue();
return e.print(os);
}

istream&
operator>>(istream& is, Dv::Util::Env& e) throw (Dv::Util::Env::Exception) {
e.read(is);
return is;
}
}
