#ifndef DV_UTIL_ENV_H
#define DV_UTIL_ENV_H
// $Id: env.h,v 1.8 2002/10/20 13:19:22 dvermeir Exp $
#include <string>
#include <stdexcept>
#include <iostream>
#include <map>
#include <list>
#include <vector>
#include <iterator>
#include <dvutil/tostring.h>
#include <dvutil/props.h>
#include "noconst.h"

/** \mainpage
\author dvermeir@vub.ac.be

<a href="../download/">Download</a>

\section install_dvssl Installation

Standard. See the INSTALL file in the top directory.
\section Dependencies Dependencies

<a href="http://tinf2.vub.ac.be/~dvermeir/software/dv/dvutil/html/index.html>dvutil</a>,

\section Introduction_dvenv Introduction

Provides a class Dv::Util::Env as a generalization of Dv::Util::Props.

*/
namespace Dv {
namespace Util {
/**
 * An Env object represents an extended ``environment''. 
 * Here, ``extended environment'' means either a string or
 * a finite mapping from string (or int) to string (or other
 * environments).
 * An Env object \a e corresponds to either
 * - a string
 * - a map from strings to environments
 * - a vector of environments
 *
 * Access is supported using conversions and index operators:
 * \code 
 * Env es, em, ev, e;
 * es = "value"; // a string environment
 * em["key"] = "value"; // a string environment
 * ev[1] = "value";
 * e["es"] = es; // nesting
 * e["key"] = em;
 * e["vec"] = ev;
 * cout << e["vec"][1]; // will print "value"
 * \endcode
 * Note however, that en environment is one of the above, combinations
 * are not allowed.
 * \code
 * Env e;
 * e["key"] = "value";
 * e[1] = "error"; // will throw exception
 * \endcode
 * For maps and vectors, read-only access is supported through
 * \a operator().
 * \code
 * cout << e("key"); // throw exception if not defined
 * e("key") = "value"; // will not compile
 * \endcode
 *
 * In general the following expressions are supported (we only show
 * three occurrences of \a [i] or \a (i) but any number of them is
 * allowed):
 * \code
 * e[i][i][i]
 * e[i][i][i] = t
 * lt = e(i)(i)(i)
 * if (e(i)(i)(i).defined()) ...
 * if (e(i)(i)(i).is_array()) ...
 * if (e(i)(i)(i).is_map()) ...
 * if (e(i)(i)(i).is_string()) ...
 * \endcode
 * where 
 * - \a i is either a string or an integer 
 * - \a t is any value of a type \a T for which Dv::Util::tostring(t) is defined
 * - \a lt is an lvalue of type \a T for which Dv::Util::fromstring(t, s) is
 * defined
 *
 * Internally, an Env object is just a pointer to a Box.
 * \sa Box
 *
 * An environment can be read from and written to a stream, in
 * a format that is illustrated below (| indicates the beginning of a line):
 * 
 * \code
 * |database::name=my-database
 * |query=select name, semester
 * |>from course
 * |answer::0::field-1=f1
 * |answer::1::field-2=f2
 * |host=mainserver
 * |# a comment
 * |age=99
 * \endcode
 *
 * The format of a single entry is
 * \code
 * entry = index '=' value \n
 * \endcode
 * where \a value is any string which is copied literally except fo
 * newline (\n) characters that are translated as "\n>".
 * 
 * Note that
 * <ul>
 * <li> Empty lines (with only white space chars) are ignored.
 * <li>Lines starting with '#' are ignored (useful for comments).
 * <li>An index has the sytax
 * \code
 * index = simple-index
 *	 | index '::' simple-index
 *	 ;
 * simple-index = name
 *              | number
 *		;
 * name = [a-zA-Z_-][a-zA-Z_-0-9]*
 * number = [0-9]+
 * \endcode
 * </ul>
 * \warning The actual input function \a operator>>(istream&,Env&)
 * is much more liberal than the above specification, but this may
 * change in future versions.
 */
class Env {
public:
  /** Exception class for Env and related classes */
  class Exception: public std::runtime_error {
  public:
    /**
     * Class name, i.e. "Env::Exception".
     */
    static const std::string NAME;
    /**
     * Constructor.
     * \param msg message to append to NAME.
     */
    Exception(const std::string& msg): std::runtime_error(NAME + ": " + msg) {}
  };

  /**
   * Create a NIL environment.
   */
  Env();
  /**
   * Create a deep copy of its argument.
   * \param e environment to be copied.
   */
  Env(const Env& e);
  /**
   * Convenience constructor. Read a fresh environment from a stream.
   * \param is stream from which environment will be read.
   * \exception Exception if bad syntax.
   */
  Env(std::istream& is) throw (Exception);
  /**
   * Convenience constructor. Read a fresh environment from a file stream.
   * \param fn name of file containing environment.
   * \exception Exception if bad syntax.
   */
  Env(const std::string& fn) throw (Exception);

  /**
   * Destructor.
   */
  ~Env();

  class iterator;


  /**
   * Return string from an underlying SBox.
   * \return string associated with SBox.
   * \exception Exception if the box underlying the environment is not
   *   an SBox.
   */
  const std::string& svalue() const throw (Exception);
  /**
   * Return non-const reference to string from an underlying SBox.
   * If there is no underlying box, an SBox will be created.
   * \return non-const reference string associated with SBox.
   * \exception Exception if the box underlying the environment is not
   *   an SBox or NIL.
   */
  std::string& slvalue() throw (Exception);

  /** 
   * Convert to any non-pointer type T for which operator>>(std::istream&,T&) is
   * defined. 
   * \return object of type T
   * \exception Exception if svalue() cannot be converted to T object.
   */
  template<typename T> 
  operator T() const throw (Exception) {
    typename noconst<T>::mutable_type t;
    if (!Dv::Util::fromstring(t,svalue()))
      throw Exception(svalue() + ": cannot convert");
    return t;
    }

  /** 
   * We refuse to convert to T*, except if T = const char.
   * \exception Exception always thrown.
   */
  template<typename T> 
  operator T*() const throw (Exception) { 
    throw Exception(svalue() + ": cannot convert fo T*");
    return 0; }

  /**
   * The exception about converting to pointers.
   */
  operator const char*() const throw (Exception) { 
    return svalue().c_str();
    }

  /**
   * Attempt to convert an Env to a string.
   * Calls Env::Slvalue().
   * \return string of associated SBox.
   * \exception Exception if underlying box is not an SBox.
   * \sa Env::slvalue()
   */
  operator const std::string&() const throw (Exception) { return svalue(); }
  /**
   * Attempt to convert an Env to a non-const reference to a string.
   * Calls Env::Slvalue().
   * \return non-const reference to string of associated SBox.
   * \exception Exception if underlying box is not an SBox.
   * \sa Env::slvalue()
   */
  operator std::string&() throw (Exception) { return slvalue(); }

  /**
   * Check whether this environment has an associated Box.
   * \return true iff there is an underlying box.
   */
  bool defined() const { return box_ != 0; }

  /**
   * Type of Box.
   */
  enum Type { NIL /** corresponds to ``no box'' */,
  	      STRING /** box contains a string */, 
	      MAP /** box contains key_string -> box map */,
	      ARRAY /** box contains [0..n[ -> box map */
	      };
  /**
   * Return type of underlying box.
   * \return type of underlying box, NIL if no box is associated with
   *   this environment.
   */
  Type type() const;

  static const char* type_s(Type t);

  /**
   * Is this environment associated with a string box?
   * \return true iff underlying box is SBox 
   */
  bool is_string() const { return type() == STRING; }
  /**
   * Is this environment associated with an array box?
   * \return true iff underlying box is ABox 
   */
  bool is_array() const { return type() == ARRAY; }
  /**
   * Is this environment associated with a map box?
   * \return true iff underlying box is MBox 
   */
  bool is_map() const { return type() == MAP; }
  
  /**
   * Return reference to a fixed constant environment without a box.
   * \return reference to an environment without a box.
   */
  static const Env& nil();

  /**
   * Return lvalue environment associated with index \a i.
   * If the there is no box, an ABox will be created.
   * \return lvalue environment associated with index \a i.
   * \exception Exception is underlying box is not an ABox or nil.
   */
  Env& operator[](int i) throw (Exception);
  /**
   * Return (const reference to) environment associated with index \a i.
   * \return (const reference to) environment associated with index \a i.
   * \exception Exception is underlying box is not an ABox.
   */
  const Env& operator()(int i) const throw (Exception);

  /**
   * Return lvalue environment associated with key string \a s.
   * If the there is no box, an MBox will be created.
   * \return reference to environment associated with key \a s.
   * \exception Exception is underlying box is not an MBox or nil.
   */
  Env& operator[](const std::string& s) throw (Exception);
  /**
   * Return environment associated with key string \a s.
   * \return (const reference to) environment associated with key \a s.
   * \exception Exception is underlying box is not an MBox.
   */
  const Env& operator()(const std::string& s) const throw (Exception);

  /** 
   * Assign a \a T object to en environment. 
   * \param t object of type \a T.
   * \return reference to \a *this.
   * \sa Env::slvalue()
   * \warnning is a no-op on Env::nil().
   */
  template<typename T>
  Env& operator=(const T& t) { 
	if (this!=&nil())
  	  slvalue() = Dv::Util::tostring(t);
  	return *this;
	}

  /**
   * Specialization of assignment for \a string.
   * \param s string to assign.
   * \return reference to \a *this.
   * \sa Env::slvalue()
   * \warnning is a no-op on Env::nil().
   */ 
  Env& operator=(const std::string& s);

  /**
   * Specialization of assignment for \a Env.
   * \param e environment to assign.
   * \return reference to \a *this.
   * \sa Env::merge()
   * \warnning is a no-op on Env::nil().
   */ 
  Env& operator=(const Env& e);

  /**
   * Clear an environment.
   * \return reference to \a *this.
   */
  Env& clear() { return *this = nil(); }
  /**
   * Merge an environment into the current one.
   * The merging proceeds as follows:
   * - if the box underlying the target is nil, the source is
   * simply copied.
   * - otherwise, if the boxes underlying the source and target are of
   * different * non-nil type, nothing is done.
   * - otherwise, if the target and source box is of type string, nothing
   * is done,
   * - otherwise, if the target and source box are of type array, the
   * target values are appended to the source values. E.g. if the
   * targets has \a i entries, \a e[0..] will be put in \a (*this)[i..]
   * - otherwise, if the target and source box are of type map, 
   * \a (*this)[key]=e[key] iff only if \a (*this)[key] is undefined.
   *
   * \param e environment to copy to \a *this.
   * \return reference to \a *this.
   */
  Env& merge(const Env& e);

  /**
   * Return number of top-level entries in underlying structure.
   * \return nr of entries in MAP or ARRAY, 1 if STRING, 0 if NIL
   */
  size_t size() const;

  /**
   * Compare after converting \a *this to T.
   * \param t object to compare with.
   * \return true iff \a static_cast<T>(*this)==t.
   * \sa Env::operator T()
   */
  template<typename T>
  bool
  operator==(const T& t) const {
  return static_cast<T>(*this) == t;
  }

  /**
   * Specialization of operator==(const T&) for \a string.
   * \param s string to compare with.
   * \return true iff \a svalue()==s.
   * \sa Env::svalue()
   */
  bool
  operator==(const std::string& s) const {
  return svalue() == s;
  }

  /**
   * Specialization of operator==(const T&) for \a Env.
   * Does a ``deep'' comparison.
   * \param e environment to compare with.
   * \return true iff all components of the environments are equal.
   */
  bool
  operator==(const Env& e) const;

  template<typename T>
  bool
  operator!=(const T& t) const {
  return ! (*this == t);
  }

  /**
   * Copy \a is to \a os, recursively replacing variables by their value
   * in \a *this.
   * A variable reference is of the form like \a ${k} (\a '$' can be
   * escaped by preceding it with \a '\') where \a k is of
   * the form
   * \code
   * i [ :: i ]+
   * \endcode
   * where each \a i is
   * either 
   * - an integer, or
   * - a string not containing any of the special
   *   character sequences \a '}', \a '*' or \a '::', or
   * - the special character \a '*' which stands for ``all integer 
   *   indices''. If \a '*' occurs, it must be the last \a i in \a k.
   *   The following illustrates its use
   *   \code
   *   ${i::*}{blabla ${a::b} more text}
   *   \endcode
   *   In the above code, if ${i} is an array, then the output
   *   will contain the result of 
   *   \code
   *   blabla ${i::0::a::b} more text
   *   ..
   *   blabla ${i::n::a::b} more text
   *   \endcode
   *   where \a 0..n are the valid indexes in \a env(i).
   *
   * \param os output stream to which text will be written.
   * \param is input stream from which text to be transformed will be
   * read.
   * \return a reference to \a *this.
   * \exception Exception if there is a syntax error. An illegal
   *  access on \a *this will be caught.
   *
   * \warning undefined variables will be left untouched. This, way
   * one can do another substitution on the result, from a differen environment.
   */
  const Env& substitute(std::ostream& os, std::istream& is) const throw (Exception);

  /**
   * Copy \a ins to \a out, recursively replacing variables by their value
   * in \a *this.
   * A variable reference is of the form like \a ${k} (\a '$' can be
   * escaped by preceding it with \a '\') where \a k is of
   * the form
   * \code
   * i [ :: i ]+
   * \endcode
   * where each \a i is
   * either 
   * - an integer, or
   * - a string not containing any of the special
   *   character sequences \a '}', \a '*' or \a '::', or
   * - the special character \a '*' which stands for ``all integer 
   *   indices''. If \a '*' occurs, it must be the last \a i in \a k.
   *   The following illustrates its use
   *   \code
   *   ${i::*}{blabla ${a::b} more text}
   *   \endcode
   *   In the above code, if ${i} is an array, then the output
   *   will contain the result of 
   *   \code
   *   blabla ${i::0::a::b} more text
   *   ..
   *   blabla ${i::n::a::b} more text
   *   \endcode
   *   where \a 0..n are the valid indexes in \a env(i).
   *
   * \param out output string to which text will be written.
   * \param ins input string from which text to be transformed will be
   * read.
   * \return a reference to \a *this.
   * \exception Exception if there is a syntax error. An illegal
   *  access on \a *this will be caught.
   *
   * \warning undefined variables will be left untouched. This, way
   * one can do another substitution on the result, from a differen environment.
   */
  const Env& substitute(std::string& out, const std::string& ins) const throw (Exception);


  /**
   * Auxiliary function for operator<<.
   * Print \a *this in the format
   * \code
   * prefix key=value
   * ..
   * \endcode
   *
   * \param os stream to print to.
   * \param prefix will be prefixed before each key
   */
  std::ostream& print(std::ostream& os, const std::string& prefix = "") const;

  /**
   */
  Env& read_key(const std::string& s) throw (Exception);
  /**
   * Decompose index of form \a "a::b::c" into vector \a [a,b,c].
   * \param index string to decompose
   * \param keys components will be appended to this vector.
   * \return true iff last component equals \a "*"
   * \warning Empty components are ignored, e.g. \a a::::b
   *   will yield \a [a,b].
   */
  static bool explode(const std::string& index, std::vector<std::string>& keys);

  /**
   * A display is a stack of pointers into an environment.
   * The bottom of the stack always contains the ``top'' 
   * environment \a e, all other entries point to a subenvironment of \a e.
   * Unlike for e.g. displays in a Pascal compiler, the stack
   * does not need to represent a path in the underlying environment.
   * Looking up an index in a display returns the first environment
   * found (search proceeds from the stack down).
   */
  class Display {
  public:
    /**
     * Initialize a display with an environment.
     * \param e root environment of display.
     */
    Display(const Env& e);
    Display(const Display& d);
    size_t size() const { return envs_.size(); }
    /**
     * \return topmost environment of display
     */
    const Env* top() const { return envs_.front(); }
    Display& push(const std::string& key);
    Display& pop();
    const Env* operator()(const std::string& key) const;
    const Env* operator()(const std::vector<std::string>& keys) const;
    Display& expand(const std::vector<std::string>& keys);
    void substitute(std::ostream& os, std::istream& is) const throw (Exception);
    std::ostream& print(std::ostream& os) const;
  private:
    std::list<const Env*> envs_;
    std::list<std::string> names_;
  };

  /**
   * Return child environment corresponding with \a key.
   * \param key int of string to use as key in \a *this.
   * \return e(key) if it exists or 0.
   */
  const Env* access(const std::string& key) const;
  /**
   * Return descendant environment corresponding with \a keys[i]...
   * \param keys vector of keys, to be traversed.
   * \param i traversal should start with keys[i]
   * \return \a e(key[i])(keys[i+1])..  if it exists or 0.
   */
  // const Env* access(const std::vector<std::string>& keys, size_t i = 0 ) const;

  /**
   * Merge environment data from is with this environment.
   * Environments written with \a operator<<(ostream&,const \a Env&)
   * can be read with this function.
   * \param is stream to read environment data from.
   * \exception Exception, e.g. due to incompatibility of data with \a e
   * or due to syntax error.
   * \sa operator<<(ostream&,const \a Env&);
   */
  Env& read(std::istream& is) throw (Dv::Util::Env::Exception);

protected:

  class SBox;
  class MBox;
  class ABox;

  /**
   * Abstract base class for boxes. 
   */
  class Box {
  public:
    /**
     * Abstract base class for (forward) iterators on containers boxes.
     */
    class iterator {
    public:
      typedef std::map<std::string,const Env*>::iterator::value_type value_type;
      virtual iterator& operator++();
      virtual value_type operator*() const;
      virtual bool equal(const iterator* i) const;
    };

    /**
     * \return Type of box.
     * \warning Uses dynamic_cast on \a *this.
     */
    Type type() const;

    /**
     * \return iterator pointing to first element (or end()) of
     * container box.
     * \exception Exception if not a map or array box.
     */
    virtual iterator* begin() const throw (Exception);
    /**
     * \return iterator pointing past the last element of
     * a container box.
     * \exception Exception if not a map or array box.
     */
    virtual iterator* end() const throw (Exception);

    /**
     * Convert SBox to string.
     * \return lvalue string of SBox.
     * \exception Exception if not a string box.
     * \sa Box::slvalue()
     */
    operator std::string&() throw (Exception) { return slvalue(); }
    /**
     * Convert SBox to constant string reference.
     * \return constant string reference of SBox.
     * \exception Exception if not a string box.
     * \sa Box::svalue()
     */
    operator const std::string&() const throw (Exception) { return svalue(); }

    /**
     * Convert SBox to string.
     * \return lvalue string of SBox.
     * \exception Exception if not a string box.
     */
    virtual const std::string& svalue() const throw (Exception);
    /**
     * Convert SBox to constant string reference.
     * \return constant string reference of SBox.
     * \exception Exception if not a string box.
     */
    virtual std::string& slvalue() throw (Exception);

    /**
     * \return non-const pointer to environment corresponding
     * to index \a i in ABox.
     * \param i index to use
     * \exception Exception if not an ABox.
     */
    virtual Env* operator[](int i) throw (Exception);
    /**
     * \return const pointer to environment corresponding
     * to index \a i in ABox.
     * \param i index to use
     * \exception Exception if not an ABox.
     */
    virtual const Env* operator()(int i) const throw (Exception);

    /**
     * \return non-const pointer to environment corresponding
     * to key \a s in MBox.
     * \param s key to use
     * \exception Exception if not an MBox.
     */
    virtual Env* operator[](const std::string& s) throw (Exception);
    /**
     * \return const pointer to environment corresponding
     * to key \a s in MBox.
     * \param s key to use
     * \exception Exception if not an MBox.
     */
    virtual const Env* operator()(const std::string& s) const throw (Exception);
    virtual ~Box();
  };

public:
  /**
   * Forward iterator class for accessing container environments.
   */
  class iterator {
  public:
    typedef Box::iterator::value_type value_type;
    iterator(Box::iterator* it): it_(it) {}
    ~iterator() { delete it_; }
    iterator& operator++() { ++*it_; return *this; }
    value_type operator*() const;
    bool operator==(const iterator& i) const { 
      return ( it_ == i.it_ ? true : it_->equal(i.it_) ); }
    bool operator!=(const iterator& i) const { return ! (*this == i); }
  private:
    Box::iterator*	it_;
  };

  iterator begin() const;
  iterator end() const;

protected:

  /** An SBox simply contains a string. */
  class SBox: public Box {
  public:
    virtual const std::string& svalue() const throw (Exception) { return s_; }
    virtual std::string& slvalue() throw (Exception) { return s_; }
  private:
    std::string s_;
  };

  /** An MBox contains a map<string,Env*> map. */
  class MBox: public Box {
  public:
    typedef std::map<std::string, Env*> Map;

    /** MBox iterator class */
    class iterator: public Box::iterator {
    public:
      iterator(Map::const_iterator it): it_(it) {}
      iterator& operator++() { ++it_; return *this; }
      value_type operator*() const { return *it_; }
      bool equal(const Box::iterator* i) const;
    private:
      Map::const_iterator it_;
    };

    iterator* begin() const throw (Exception) { return new iterator(m_.begin()); }
    iterator* end() const throw (Exception) { return new iterator(m_.end()); }

    MBox();
    ~MBox();

    /**
     * \return number of elements in underlying map.
     */
    size_t size() const { return m_.size(); }
    /**
     * A new Env will be allocated if none exists.
     * \param s key to use
     * \return pointer to Env corresponding to key.
     */
    Env* operator[](const std::string& s) throw (Exception);
    /**
     * A new Env will be allocated if none exists.
     * \param s key to use
     * \return pointer to Env corresponding to key.
     */
    const Env* operator()(const std::string& s) const throw (Exception);
  private:
    Map m_;
  };

  /** An ABox contains a \a vector<Env*> array . */
  class ABox: public Box {
  public:
    typedef std::vector<Env*> Vector;

    /** ABox iterator class. */
    class iterator: public Box::iterator {
    public:
      iterator(Vector::const_iterator it, const Vector* v);
      iterator& operator++() { ++it_; return *this; }
      value_type operator*() const;
      bool equal(const Box::iterator* i) const;
    private:
      Vector::const_iterator it_;
      const Vector* const v_;
    };

    iterator* begin() const throw (Exception) { return new iterator(a_.begin(), &a_); }
    iterator* end() const throw (Exception) { return new iterator(a_.end(), &a_); }

    /**
     * Constructor
     * \param size elements will be allocated, initialized to 0.
     */
    ABox(size_t size);
    ~ABox();
    size_t size() const { return a_.size(); }
    Env* operator[](int i) throw (Exception);
    const Env* operator()(int i) const throw (Exception);
  private:
    Vector a_;
  };


private:
  Box*	box_;
};

}}

namespace std {
/**
 * Write an environment to a stream.
 * Environments written by this function
 * can be read back by \a operator>>(istream&,Env&).
 * \param os stream to write environment data to.
 * \param e environment to write.
 * \sa operator>>(istream&,const \a Env&);
 */
std::ostream&
operator<<(std::ostream& os, const Dv::Util::Env& e);
/**
 * Merge environment data from is with e.
 * Environments written with \a operator<<(ostream&,const \a Env&)
 * can be read with this function.
 * \param is stream to read environment data from.
 * \param e target environment to merge with.
 * \exception Exception, e.g. due to incompatibility of data with \a e
 * or due to syntax error.
 * \sa operator<<(ostream&,const \a Env&);
 */
std::istream&
operator>>(std::istream& is, Dv::Util::Env& e) throw (Dv::Util::Env::Exception);

inline std::ostream&
operator<<(std::ostream& os, const Dv::Util::Env::Display& d) {
return d.print(os);
}
}
#endif
